#include "subdivision.h"

Subdivision::Subdivision(QOpenGLFunctions_4_3_Core *f)
{
    this->f =f;
}

Subdivision::~Subdivision()
{
    delete edgeShader;
}

void Subdivision::init() {
    QString source = QLatin1String(":/subdivision-edge.compute");
    edgeShader = initComputeShaderProgram(source);

    source = QLatin1String(":/subdivision-vertex.compute");
    vertexShader = initComputeShaderProgram(source);
}

QOpenGLShaderProgram *Subdivision::initComputeShaderProgram(QString &source){
    qDebug()<<"Compiling compute shader ...";
    QOpenGLShader *computeShader = new QOpenGLShader(QOpenGLShader::Compute);
    if(!computeShader->compileSourceFile(source)){
        qCritical()<<"Compute Shader"<<source<<"failed"<<computeShader->log();
        exit(5);
    }

    qDebug()<<"Adding compute shader ...";
    QOpenGLShaderProgram *shader = new QOpenGLShaderProgram();
    shader->addShader(computeShader);
    qDebug()<<"Linking compute shader ...";
    if(!shader->link()){
        qCritical()<<"Linking compute shader failed:"<<shader->log();
        exit(5);
    }
    qDebug()<<"Linking compute shader done";

    return shader;
}

Mesh *Subdivision::subdivide(Mesh *mesh) {
    // For now we only look at the first mesh entry
    Mesh::Node root = mesh->getRootNode();

    int first_mesh_index = -1;
    if (!root.getFirstMeshIndex(first_mesh_index)) {
        qCritical()<<"No mesh found, aborting subdivision";
        return NULL;
    }

    Mesh::MeshEntry *current_mesh = mesh->getMeshEntry(first_mesh_index);
    Input input;
    input.vb_handle = current_mesh->VB_handle;
    input.vertex_buffer = current_mesh->getVertexBuffer();
    input.index_buffer = current_mesh->getIndexBuffer();

    Tables tables = precomputeTables(input);
    Result result = runShader(input, tables);

    return NULL;
}

/**
 * @brief Subdivision::precomputeTables
 * @param mesh
 * @param edgeIndices_base Wird gefüllt mit 4er-Arrays, die die indices in den alten Vertexbuffer enthalten, von denen der jeweilige edge-point abhängt
 * @param vertexIndices Wird gefüllt mit Folgen von ints, die die indices in den alten Vertexbuffer enthalten, von denen die neuen vertex-points abhängen
 * @param vertexIndicesOffsets enthält offsets in vertexIndices
 * @param newIndexBuffer Wird gefüllt mit neuem Indexbuffer. Der Indexbuffer geht davon aus, dass im neuen Vertex-Buffer
 * erst die Vertex-points (in der selben Reihenfolge wie im ursprünglichen Vertex-Buffer) und dann die Edgepoints kommen.
 *
 * Einträge in neuem Vertex-buffer: Verschobene Knoten aus altem Vertex-Buffer :: edge-vertices erstellt aus edgeIndices_base.
 */
Subdivision::Tables Subdivision::precomputeTables(Input input) {
    Tables tables;

    QVector<unsigned int> ib = input.index_buffer;
    qDebug()<<"Index Buffer: "<<ib;

    QVector<Vertex> vb = input.vertex_buffer;
    //qDebug()<<"Vertex Buffer: "<<vb;

//compute edge table
    //Format: first two entries: edge vertices. last two entries: distant vertices.

    unsigned int nib_offset = vb.length();
    for (int i = 0; i < ib.length(); i+=3){
        //schaue alle dreiecke an

        Vertex x, y, z;
        unsigned int x_i = ib[i];
        unsigned int y_i = ib[i+1];
        unsigned int z_i = ib[i+2];

        //get vertices x,y,z from vertex buffer
        x = vb[x_i];//todo maybe check array length
        y = vb[y_i];
        z = vb[z_i];


        /*
         * Find edge xy in all other triangles, add to edge_indices
         * Find edge yz
         * Find edge zx
         * - We assume that we have a closed surface, i.e. each triangle has exactly one adjacent triangle at each edge
         * - Each edge point will be computed twice. This allows us to implicitly know the adjacent points.
         */

        //find indices for edge point on xy-edge
        for (int j = 0; j < ib.length(); j+= 3){
            if (j != i){
                //search all following triangles for common edge
                Vertex a,b,c;
                unsigned int a_i, b_i, c_i;
                a_i = ib[j];
                b_i = ib[j+1];
                c_i = ib[j+2];

                a = vb[a_i];
                b = vb[b_i];
                c = vb[c_i];

                //get vertices for a,b,c from vertex buffer
                //comparisons: xy = ba, cb, ac.
                //yz = ba, cb, ac
                //zx = ba, cb, ac
                //TODO if oder else if?
                if (x.samePos(a) && y.samePos(c)){
                    tables.edge_indices.push_back(x_i);
                    tables.edge_indices.push_back(y_i);
                    tables.edge_indices.push_back(z_i);
                    tables.edge_indices.push_back(b_i);
                }
                if (x.samePos(b) && y.samePos(a)){
                    tables.edge_indices.push_back(x_i);
                    tables.edge_indices.push_back(y_i);
                    tables.edge_indices.push_back(z_i);
                    tables.edge_indices.push_back(c_i);
                }
                if (x.samePos(c) && y.samePos(b)){
                    tables.edge_indices.push_back(x_i);
                    tables.edge_indices.push_back(y_i);
                    tables.edge_indices.push_back(z_i);
                    tables.edge_indices.push_back(a_i);
                }
            }
        }

        //find indices for edge point on yz-edge
        for (int j = 0; j < ib.length(); j+= 3){
            if (j != i){
                //search all following triangles for common edge
                Vertex a,b,c;
                unsigned int a_i, b_i, c_i;
                a_i = ib[j];
                b_i = ib[j+1];
                c_i = ib[j+2];

                a = vb[a_i];
                b = vb[b_i];
                c = vb[c_i];

                //yz = ba, cb, ac
                if (y.samePos(a) && z.samePos(c)){
                    tables.edge_indices.push_back(y_i);
                    tables.edge_indices.push_back(z_i);
                    tables.edge_indices.push_back(x_i);
                    tables.edge_indices.push_back(b_i);
                }
                if (y.samePos(b) && z.samePos(a)){
                    tables.edge_indices.push_back(y_i);
                    tables.edge_indices.push_back(z_i);
                    tables.edge_indices.push_back(x_i);
                    tables.edge_indices.push_back(c_i);
                }
                if (y.samePos(c) && z.samePos(b)){
                    tables.edge_indices.push_back(y_i);
                    tables.edge_indices.push_back(z_i);
                    tables.edge_indices.push_back(x_i);
                    tables.edge_indices.push_back(a_i);
                }
            }
        }

        //find indices for edge point on xy-edge
        for (int j = 0; j < ib.length(); j+= 3){
            if (j != i){
                //search all following triangles for common edge
                Vertex a,b,c;
                unsigned int a_i, b_i, c_i;
                a_i = ib[j];
                b_i = ib[j+1];
                c_i = ib[j+2];

                a = vb[a_i];
                b = vb[b_i];
                c = vb[c_i];


                if (x.samePos(a) && z.samePos(b)){
                    tables.edge_indices.push_back(x_i);
                    tables.edge_indices.push_back(z_i);
                    tables.edge_indices.push_back(y_i);
                    tables.edge_indices.push_back(c_i);
                }
                if (x.samePos(b) && z.samePos(c)){
                    tables.edge_indices.push_back(x_i);
                    tables.edge_indices.push_back(z_i);
                    tables.edge_indices.push_back(y_i);
                    tables.edge_indices.push_back(a_i);
                }
                if (x.samePos(c) && z.samePos(a)){
                    tables.edge_indices.push_back(x_i);
                    tables.edge_indices.push_back(z_i);
                    tables.edge_indices.push_back(y_i);
                    tables.edge_indices.push_back(b_i);
                }
            }
        }//quadratische Laufzeit ftw

        //add indices to new index buffer
            tables.index_buffer.push_back(ib[i]);
            tables.index_buffer.push_back(nib_offset + i);
            tables.index_buffer.push_back(nib_offset + i + 2);

            tables.index_buffer.push_back(ib[i+1]);
            tables.index_buffer.push_back(nib_offset + i + 1);
            tables.index_buffer.push_back(nib_offset + i);

            tables.index_buffer.push_back(ib[i+2]);
            tables.index_buffer.push_back(nib_offset + i + 2);
            tables.index_buffer.push_back(nib_offset + i + 1);

            tables.index_buffer.push_back(nib_offset + i);
            tables.index_buffer.push_back(nib_offset + i + 1);
            tables.index_buffer.push_back(nib_offset + i + 2);
       //Wichtig: Wir gehen davon aus, dass wir geschlossene Oberflächen haben, dh für jede Kante von einem Dreieck wird eine passende Kante bei einem anderen Dreieck gefunden.
    }//for each index in indexbuffer

    qDebug()<<"Done with edge table. "<<tables.edge_indices.length()<<" edges found. Table: "<<tables.edge_indices;
    if (true){
        qDebug()<<"Table (long version):";
        for (int i = 0; i < tables.edge_indices.length(); i++){
            qDebug()<<"blub:"<<tables.edge_indices[i];
        }
    }

//compute vertex table
    //Format: First entry: original vertex. Other entries: adjacent vertices.
    QVector<QVector<unsigned int> > duplicates;//for debugging
    unsigned int offset = 0;
    for (unsigned int i = 0; i < vb.length(); i++){
        //hat evtl viel redundanz

        QVector<Vertex> adj_v;//helfer
        tables.vertex_offsets.push_back(offset);
        tables.vertex_indices.push_back(i);
        offset++;

        Vertex originalVertex = vb[i];

        QVector<unsigned int> d;

        for (int j = 0; j < ib.length(); j++){

            //suche verweise auf vertex im indexbuffer
            if (vb[ib[j]].samePos(originalVertex)){
                d.push_back(j);
                unsigned int i1, i2; //indices for neighbour vertices
                if(j%3==0){
                    i1 = ib[j+1];
                    i2 = ib[j+2];
                } else if(j%3==1){
                    i1 = ib[j-1];
                    i2 = ib[j+1];
                } else{
                    i1 = ib[j-2];
                    i2 = ib[j-1];
                }

                Vertex v1,v2; //neighbour vertices;
                v1 = vb[i1];
                v2 = vb[i2];
                //check if vertices where already used to influence this vertex
                //Note: Can't use contain. Contain uses equal (would have to implement equal, but we're only interested in the position here.
                bool found1 = false;
                bool found2 = false;
                for (int k = 0; k < adj_v.length(); k++){
                    if (adj_v[k].samePos(v1)) found1 = true;
                    if(adj_v[k].samePos(v2)) found2 = true;
                }
                if (!found1){
                    adj_v.push_back(v1);
                    tables.vertex_indices.push_back(i1);
                    offset++;
                }
                if (!found2){
                    adj_v.push_back(v2);
                    tables.vertex_indices.push_back(i2);
                    offset++;
                }
                //if one adjacent vertex is referenced by multiple different indices, only its first index will appear in the adjacent_indices list.1
            }
        }
        if (!duplicates.contains(d)) duplicates.push_back(d);
    }

    qDebug()<<"Done with vertex index table. ";

    tables.vertex_offsets.push_back(offset);

    qDebug()<<"Duplicates: ";
    for (int i = 0; i < duplicates.length(); i++){
        qDebug()<<duplicates[i];
    }

    return tables;
}

void Subdivision::extractPatches(QVector<unsigned int> index_buffer, QVector<Vertex> vertex_buffer, QVector<Subdivision::Patch> &patches){
    patches.clear();

    for (unsigned int i = 0; i < index_buffer.length()-2; i += 3){
        Vertex vx = vertex_buffer[index_buffer[i]];
        Vertex vy = vertex_buffer[index_buffer[i+1]];
        Vertex vz = vertex_buffer[index_buffer[i+2]];

        int countx = 0;
        int county = 0;
        int countz = 0;
        for(unsigned int j = 0; j < index_buffer.length(); j++){
            if (j != i && j != i+1 && j != i+2){
                if (vx.samePos(vertex_buffer[index_buffer[j]])){
                    countx++;
                }
                if (vy.samePos(vertex_buffer[index_buffer[j]])){
                    county++;
                }
                if (vz.samePos(vertex_buffer[index_buffer[j]])){
                    countz++;
                }
            }
        }
        if (countx != 5 || county != 5 || countz != 5){
            //triangle is irregular
        }

    }


}

Subdivision::Result Subdivision::runShader(Input input, Tables &tables) {
    qDebug()<<"Running compute shader";

    Result result;

    // Create an input buffer for the vertex indices
    GLuint vertex_indices_handle;
    f->glGenBuffers(1, &vertex_indices_handle);
    f->glBindBuffer(GL_SHADER_STORAGE_BUFFER, vertex_indices_handle);
    f->glBufferData(GL_SHADER_STORAGE_BUFFER, tables.vertex_indices.size() * sizeof(GLuint), tables.vertex_indices.data(), GL_DYNAMIC_DRAW);

    // Create an input buffer for the vertex offsets
    GLuint vertex_offsets_handle;
    f->glGenBuffers(1, &vertex_offsets_handle);
    f->glBindBuffer(GL_SHADER_STORAGE_BUFFER, vertex_offsets_handle);
    f->glBufferData(GL_SHADER_STORAGE_BUFFER, tables.vertex_offsets.size() * sizeof(GLuint), tables.vertex_offsets.data(), GL_DYNAMIC_DRAW);

    // Create an input buffer for the edge indices
    GLuint edge_indices_handle;
    f->glGenBuffers(1, &edge_indices_handle);
    f->glBindBuffer(GL_SHADER_STORAGE_BUFFER, edge_indices_handle);
    f->glBufferData(GL_SHADER_STORAGE_BUFFER, tables.edge_indices.size() * sizeof(GLuint), tables.edge_indices.data(), GL_DYNAMIC_DRAW);

    // Create an output buffer
    GLuint output_handle;
    f->glGenBuffers(1, &output_handle);
    f->glBindBuffer(GL_SHADER_STORAGE_BUFFER, output_handle);
    f->glBufferData(GL_SHADER_STORAGE_BUFFER, (input.vertex_buffer.size() + tables.edge_indices.size() / 4) * sizeof(Vertex), NULL, GL_DYNAMIC_DRAW);
    result.vb_handle = output_handle;

    f->glBindBuffer(GL_SHADER_STORAGE_BUFFER, 0);

    runVertexShader(input.vertex_buffer.size(), input.vb_handle, vertex_indices_handle, vertex_offsets_handle, output_handle);
    int edgeOffset = input.vertex_buffer.size();
    runEdgeShader(tables.edge_indices.size() / 4, input.vb_handle, edge_indices_handle, output_handle, edgeOffset);

    // Map the output buffer so we can read the results
    f->glBindBuffer(GL_SHADER_STORAGE_BUFFER, output_handle);
    Vertex *ptr;
    ptr = (Vertex *) f->glMapBuffer(GL_SHADER_STORAGE_BUFFER, GL_WRITE_ONLY);
    qDebug() << "New vertices:";
    for (int i = 0; i < input.vertex_buffer.size(); i++) {
        qDebug() << ptr[i].pos;
	result.vertex_buffer.push_back(ptr[i]);
    }
    qDebug() << "New edge points:";
    for (int i = 0; i < tables.edge_indices.size() / 4; i++) {
        qDebug() << ptr[edgeOffset + i].pos;
	result.vertex_buffer.push_back(ptr[edgeOffset + i]);
    }
    f->glUnmapBuffer(GL_SHADER_STORAGE_BUFFER);

    // Delete the buffers the free the resources
    f->glDeleteBuffers(1, &vertex_indices_handle);
    f->glDeleteBuffers(1, &vertex_offsets_handle);
    f->glDeleteBuffers(1, &edge_indices_handle);

    return result;
}

void Subdivision::runVertexShader(GLuint size, GLuint vb_handle, GLuint vertex_indices_handle, GLuint vertex_offsets_handle, GLuint output_handle) {
    vertexShader->bind();

    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 0, vb_handle);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 1, vertex_indices_handle);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 2, vertex_offsets_handle);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 3, output_handle);

    // Run the shader
    f->glDispatchCompute(size, 1, 1);

    // Wait for the shader to complete and the data to be written back to the global memory
    f->glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);

    // Unbind the buffers from the shader
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 0, 0);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 1, 0);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 2, 0);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 3, 0);

    vertexShader->release();
}

void Subdivision::runEdgeShader(GLuint size, GLuint vb_handle, GLuint edge_indices_handle, GLuint output_handle, GLuint offset) {
    edgeShader->bind();

    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 0, vb_handle);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 1, edge_indices_handle);
    f->glBindBufferRange(GL_SHADER_STORAGE_BUFFER, 2, output_handle, offset * sizeof(Vertex), size * sizeof(Vertex));

    // Run the shader
    f->glDispatchCompute(size, 1, 1);

    // Wait for the shader to complete and the data to be written back to the global memory
    f->glMemoryBarrier(GL_SHADER_STORAGE_BARRIER_BIT);

    // Unbind the buffers from the shader
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 0, 0);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 1, 0);
    f->glBindBufferBase(GL_SHADER_STORAGE_BUFFER, 2, 0);

    edgeShader->release();
}
